/*
 * Copyright 2012 Adaptrex, LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.adaptrex.core.ext.data;

import com.adaptrex.core.Adaptrex;
import com.adaptrex.core.persistence.AdaptrexPersistence;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonInclude;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.List;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class ModelDefinition {

	@JsonInclude public final String extend = "Ext.data.Model";
	
	private static AdaptrexPersistence persistence = Adaptrex.getPersistence();
	
	private Logger log = LoggerFactory.getLogger(ModelDefinition.class);

	private Config					config;
	private Class<?>				clazz;
	private boolean					isRoot;
	private String					modelName;
	private String					simpleModelName;
	
	private List<FieldDefinition>	fields = new ArrayList<FieldDefinition> ();
	private String					idProperty;
	
	private List<Association>		associations = new ArrayList<Association>();

	public ModelDefinition() {}
	
	
	/*
	 * Root Model Definition.  This is the model definition created directly
	 * by our applications.  It may contain a list of associations which each
	 * need their own model definition's defined.
	 */
	public ModelDefinition(Config config) {
		this.config = config;
		this.clazz = config.getClazz();
		this.simpleModelName = config.getModelName();
		this.modelName = config.getNamespace() + ".model." + config.getModelName();
		this.isRoot = true;
		this.initFields();
	}
	
	/*
	 * Associated Model Definition.  Any models related to an association need
	 * to be automatically created and attached to the association.  In general,
	 * this constructor never needs to be called outside the framework
	 */
	public ModelDefinition(Association association) {
		this.config = association.getConfig();
		this.clazz = association.getAssociatedModelClazz();
		this.simpleModelName = association.getModelName();
		this.modelName = simpleModelName;
		this.isRoot = false;
		this.initFields();
		
		/*
		 * hasMany associations need a corresponding belongsTo association
		 */
		if (association.getType().equals("hasMany")) {
			try {
				ModelDefinition ownerModelDefinition = association.getOwnerModelDefinition();
				String returnAssocModelName = ownerModelDefinition.getModelName();
				String returnAssocName = ownerModelDefinition.getClazz().getSimpleName().toLowerCase();
				this.associations.add(new Association("belongsTo", returnAssocModelName, returnAssocName));
			} catch (ClassNotFoundException e) {
				log.warn("Error", e);
			}
		}
	}
	
	
	/*
	 * Set up the fields for this model
	 */
	private void initFields() {
		try {
			List<String> includes = this.config.getIncludes();
			List<String> excludes = this.config.getExcludes();
			
			List<String> entityIncludes	= new ArrayList<String>();
			List<String> entityExcludes	= new ArrayList<String>();

			if (isRoot) {
				/*
				 * Loop each include/exclude/join config and determine if they
				 * are set for the root entity.
				 */
				for (String incl : this.config.getIncludes()) {
					if (!incl.contains(".")) entityIncludes.add(incl);
				}
				for (String excl : this.config.getExcludes()) {
					if (!excl.contains(".")) entityExcludes.add(excl);
				}
			} else {
				/*
				 * Loop each include/exclude/join config and determine if they are set for 
				 * the current entity at this position in the tree.  If the full path to a 
				 * specific field on this entity is set, we want to include it in the entity 
				 * specific config.  We only need the field portion when testing against 
				 * the current entity so that gets split out before adding to the list
				 */
				for (String incl : this.config.getIncludes()) {
					if (incl.contains(simpleModelName + ".")) {
						entityIncludes.add(incl.split("\\.")[1]);
					}
				}
				
				for (String excl : this.config.getExcludes()) {
					if (excl.contains(simpleModelName + ".")) {
						entityExcludes.add(excl.split("\\.")[1]);
					}
				}
			}			

			
			
			/*
			 * Create and add the fields for this model
			 */
			for (java.lang.reflect.Field field : this.clazz.getDeclaredFields()) {
				
				/*
				 * Static fields don't get returned
				 */
				if (Modifier.isStatic(field.getModifiers())) continue;
				
				
				/*
				 * Get the field name
				 */
				String fieldName = field.getName();
				
				

				/*
				 * If this field is an id, identify it as such in the model and create it
				 */
				if (persistence.isIdField(this.clazz, fieldName)) {
					this.fields.add(new FieldDefinition(field));
					this.idProperty = fieldName;
					continue;
				}				
				
				/*
				 * We don't create oneToMany columns explicitly
				 */
				if (persistence.isOneToMany(this.clazz, fieldName)) continue;
				
				
				/*
				 * Include or exclude?
				 */
				if (isRoot) {
					if (includes.size() > 0 && 
							(!includes.contains("*") && !includes.contains(fieldName))) continue;
					if (excludes.contains("*") || excludes.contains(fieldName)) continue;
					
				} else {
					if (includes.size() > 0) {
						if (entityIncludes.size() == 0 ||
								(!entityIncludes.contains("*") && !entityIncludes.contains(fieldName))) continue;
					}

					if (excludes.size() > 0) {
						if (entityExcludes.contains("*") || entityExcludes.contains(fieldName)) continue;
					}
				}
				
				/*
				 * Add many to one id field
				 */
				if (persistence.isManyToOne(this.clazz, fieldName)) {
					this.fields.add(new FieldDefinition(fieldName + "Id", "int", true));
				}
				
				this.fields.add(new FieldDefinition(field));
			}
			
			
			/*
			 * Add this model's associations
			 */
			for (String associationName : config.getAssociationEntityNames()) {
				if (isRoot && associationName.contains(".")) continue;
				if (!isRoot && !associationName.contains(".")) continue;

				if (associationName.contains(".") && 
						!associationName.startsWith(this.simpleModelName + ".")) continue;
				
				try {
					this.associations.add(new Association(this, associationName));
				} catch (ClassNotFoundException e) {}
			}
		} catch (Exception e) {
			log.warn("Error", e);
		}
	}
	
	
	@JsonIgnore
	protected Config getConfig() {
		return config;
	}
	
	@JsonIgnore
	protected Class<?> getClazz() {
		return this.clazz;
	}
	
	@JsonIgnore
	protected String getModelName() {
		return this.modelName;
	}	

	
	
	/*
	 * This is only different than modelName on top level models
	 * where there is a fully Ext namespaced model name set
	 */
	@JsonInclude
	public List<FieldDefinition> getFields() {
		return this.fields;
	}

	@JsonInclude(JsonInclude.Include.NON_NULL)
	public String getIdProperty() {
		return this.idProperty;
	}

	@JsonInclude(JsonInclude.Include.NON_EMPTY)
	public List<Association> getAssociations() {
		return associations;
	}
	
	@JsonInclude
	public RestProxy getProxy() {
		RestProxy proxy = (RestProxy) this.config.getProxy();
		if (isRoot) return proxy;

		/*
		 * If this model is an associated model, we need to create a custom proxy
		 */
		String url = proxy.getUrl().split("/rest/")[0] + "/rest/" + this.getClazz().getSimpleName().toLowerCase();
		RestProxy newProxy = new RestProxy(url, config);
		newProxy.getExtraParams().put("modelName", this.simpleModelName);
		return newProxy;
	}

}
